
import qt
import qtcanvas
import qwave2
import os
import sys
import math
import copy

__all__ = ["Waveform", "WaveformWithResizableSelection"]


class MagControl:
    """ Controls the waveform amplifier.
    """
    def __init__(self, mag, waveform, vruler):
        #self.mag = mag
        self.waveform = waveform
        self.vruler = vruler
        mag.setRange(-50,50)
        mag.setValue(0)
        
        mag.connect(mag,qt.SIGNAL("valueChanged(int)"),self.setAmplitude)

    def setAmplitude(self, v):
        self.waveform.setAmplitudeRatio(math.pow(1.35,-v/10.0))
        self.vruler.redraw()
        self.vruler.update()
        
class OutputControl:
    """ Controls 1) the relative volume, and 2) output channels.
    """
    def __init__(self, player, sndfile, ch,
                 volSlide, leftCheck, rightCheck):
        self.player = player
        self.sndfile = sndfile
        self.channel = ch
        self.left = leftCheck
        self.right = rightCheck
        self.vol = volSlide
        
        if ch % 2 == 0:
            leftCheck.setChecked(True)
            rightCheck.setChecked(False)
            player.setOutputChannel(sndfile, ch, 0)
        else:
            leftCheck.setChecked(False)
            rightCheck.setChecked(True)
            player.setOutputChannel(sndfile, ch, 1)
        volSlide.setRange(0,200)
        volSlide.setValue(100)
        player.setWeight(sndfile, ch, 1.0)

        leftCheck.connect(leftCheck,qt.SIGNAL("toggled(bool)"),self.setOutLeft)
        rightCheck.connect(rightCheck,qt.SIGNAL("toggled(bool)"),self.setOutRight)
        volSlide.connect(volSlide,qt.SIGNAL("valueChanged(int)"),self.setVol)
        
    def setVol(self, v):
        if self.left.isChecked() or self.right.isChecked():
            self.player.setWeight(self.sndfile, self.channel, (200.0-v)/100.0)

    def setOutLeft(self, v):
        if v:
            self.player.setWeight(self.sndfile, self.channel, (200.0-self.vol.value())/100.0)
            if self.right.isChecked():
                self.player.setOutputChannel(self.sndfile,self.channel,-1)
            else:
                self.player.setOutputChannel(self.sndfile,self.channel,0)
        else:
            if self.right.isChecked():
                self.player.setWeight(self.sndfile, self.channel, (200.0-self.vol.value())/100.0)
                self.player.setOutputChannel(self.sndfile,self.channel,1)
            else:
                self.player.setWeight(self.sndfile, self.channel, 0.0)

    def setOutRight(self, v):
        if v:
            self.player.setWeight(self.sndfile, self.channel, (200.0-self.vol.value())/100.0)
            if self.left.isChecked():
                self.player.setOutputChannel(self.sndfile,self.channel,-1)
            else:
                self.player.setOutputChannel(self.sndfile,self.channel,1)
        else:
            if self.left.isChecked():
                self.player.setWeight(self.sndfile, self.channel, (200.0-self.vol.value())/100.0)
                self.player.setOutputChannel(self.sndfile,self.channel,0)
            else:
                self.player.setWeight(self.sndfile, self.channel, 0.0)
        
class WaveformWithResizableSelection(qwave2.Waveform):
    """  WaveformWithResizableSelection combines WaveformSelectionProxy
    and WaveformCursorProxy to support a resizable selection.

    In terms of programming model, this class allows programmers to
    avoid dealing with WaveformSelectionProxy and WaveformCursorProxy
    directly.  Instead, it creates an illusion that this waveform
    class defines the selection, cursor and player cursor in it and
    manages them by default.
    """
    
    def __init__(self, sndfile, ch, beg, dur,
                 selProxy, curProxy, parent=None):
        qwave2.Waveform.__init__(self, sndfile, ch, beg, dur, parent)
        selProxy.registerWaveform(self)
        curProxy.registerWaveform(self)
        self.selectionProxy = selProxy
        self.cursorProxy = curProxy
        self.resizing = False
        self.resizeModifier = qt.Qt.ShiftButton
        self.keyState = None

    def __del__(self):
        self.cursorProxy.unregisterWaveform(self)
        self.selectionProxy.unregisterWaveform(self)
        
    def isSelectionResizing(self):
        """ Tells whether the selection is being resized.

        @return: Whether the selection is being resized.
        @rtype: Boolean
        """
        return self.resizing

    def setSelectionResizing(self, v):
        self.resizing = v
        
    def setResizeModifier(self, modifier):
        """ Set resize modifier.

        Resize modifier is the key that activate resizing. User should
        hold this key while moving the mouse in order to resize the
        selection.

        @param modifier: The modifier key.
        @type modifier: Qt.ButtonState
        """
        self.resizeModifier = modifier

    def getKeyState(self):
        """
        @return: The button state at the time of the last mouse event
        on the waveform.  It's actually the value of Event::state()
        call.
        @type: Qt.ButtonState
        """
        return self.keyState
        
    def getResizeModifier(self):
        """ Resize modifier is a modifier key, e.g. "Ctrl", "Shift",
        etc.  To resize a selected region on the waveform, the user
        can point the mouse to an edge of the region while holding
        the modifier key.

        @return: Resize modifier.
        @rtype: Qt.ButtonState
        """
        return self.resizeModifier

    def setAmplitude(self, v):
        self.setAmplitudeRatio(math.pow(1.35,-v/10.0))

    def contentsMousePressEvent(self, e):
        self.keyState = e.state()
        if self.keyState == self.resizeModifier:
            if self.selectionProxy.getSelectedWaveform() == self:
                x = self.cursorProxy.getCursorPositionSeconds()
                beg = self.selectionProxy.getBeginSeconds()
                end = beg + self.selectionProxy.getWidthSeconds()
                delta = self.getSecondsPerPixel() * 2.0
                if abs(x-beg) < delta:
                    self.resizing = True
                    self.selectionProxy.expandSelectionBegin(self, beg)
                    self.setCursor(qt.QCursor(qt.Qt.SizeHorCursor))
                elif abs(x-end) < delta:
                    self.resizing = True
                    self.selectionProxy.expandSelectionEnd(self, end)
                    self.setCursor(qt.QCursor(qt.Qt.SizeHorCursor))
        else:
            self.resizing = False
            qwave2.Waveform.contentsMousePressEvent(self, e)
            self.setCursor(qt.QCursor(qt.Qt.ArrowCursor))
            
    def contentsMouseReleaseEvent(self, e):
        self.keyState = e.state()
        qwave2.Waveform.contentsMouseReleaseEvent(self, e)
        self.setCursor(qt.QCursor(qt.Qt.ArrowCursor))
        if self.resizing:
            self.resizing = False
                
    def contentsMouseMoveEvent(self, e):
        self.keyState = e.state()
        qwave2.Waveform.contentsMouseMoveEvent(self, e)

        if self.keyState == self.resizeModifier and \
           e.button() == qt.Qt.NoButton and \
           self.selectionProxy.getSelectedWaveform() == self:
            x = self.cursorProxy.getCursorPositionSeconds()
            beg = self.selectionProxy.getBeginSeconds()
            end = beg + self.selectionProxy.getWidthSeconds()
            delta = self.getSecondsPerPixel() * 2.0
            if min(abs(x-beg),abs(x-end)) < delta:
                self.setCursor(qt.QCursor(qt.Qt.SizeHorCursor))
            else:
                self.setCursor(qt.QCursor(qt.Qt.ArrowCursor))
        elif self.keyState & qt.Qt.LeftButton:
            self.setCursor(qt.QCursor(qt.Qt.SizeHorCursor))
            x = self.cursorProxy.getCursorPositionSeconds()
            beg = self.getBeginSeconds()
            end = beg + self.getWidthSeconds()
            if x < beg and x >= 0.0:
                self.display(x)
            elif x > end:
                self.display(beg + x - end)
        else:
            self.setCursor(qt.QCursor(qt.Qt.ArrowCursor))

class MyHBox(qt.QWidget):
    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.layout = qt.QHBoxLayout(self)
        self.layout.setSpacing(1)
        self.w = []

    def __getitem__(self, i):
        return self.w[i]

    def __len__(self):
        return len(self.w)

    def setStretchFactor(self, i, f):
        self.layout.setStretchFactor(self.w[i],f)
        
    def initialize(self, w1=None, w2=None, w3=None, w4=None):
        if self.w: return   # you can call this only once
        if w1 is None:
            w1 = qt.QWidget(self)
            #w1.setFixedWidth(0)
            #w1.setFixedHeight(0)
        if w2 is None:
            w2 = qt.QWidget(self)
            #w2.setFixedWidth(0)
            w2.setFixedHeight(0)
        if w3 is None:
            w3 = qt.QWidget(self)
            #w3.setFixedWidth(0)
            w3.setFixedHeight(0)
        if w4 is None:
            w4 = qt.QWidget(self)
            #w4.setFixedWidth(0)
            w4.setFixedHeight(0)
        self.layout.addWidget(w1)
        self.layout.addWidget(w2)
        self.layout.addWidget(w3)
        self.layout.addWidget(w4)
        self.w.append(w1)
        self.w.append(w2)
        self.w.append(w3)
        self.w.append(w4)
        self.layout.setStretchFactor(w1,0)
        self.layout.setStretchFactor(w2,1)
        self.layout.setStretchFactor(w3,0)
        self.layout.setStretchFactor(w4,0)
        self.show()


class WaveformGrid(qt.QWidget):
    """ WaveformGrid is a grid containing a stack of waveforms.

    It has a ruler at the top row, and you can add as many waveforms as
    you want from the second row.  Column 1 is the vertical ruler,
    column 3 is a slider that is used to magnify the waveform, and
    column 4 is a slider for volume control and a check box for mutt.
    But, you don't need to worry about these, they are created and
    managed automatically.
    """
    def __init__(self, parent, waveformCls=WaveformWithResizableSelection):
        """
        @param parent:
        @type parent: Waveform
        """
        qt.QWidget.__init__(self, parent)
        self.waveformCls = waveformCls
        self.grid = qt.QVBoxLayout(self)
        self.grid.setSpacing(1)
        box1 = MyHBox(self)
        box2 = MyHBox(self)
        self.ruler1 = qwave2.WaveformRuler(True,box1)
        self.ruler2 = qwave2.WaveformRuler(False,box2)
        box1.initialize(w2=self.ruler1)
        box2.initialize(w2=self.ruler2)
        self.grid.addWidget(box1)
        self.grid.addWidget(box2)
        self.rows = [box1,box2]
        self.selection = qwave2.WaveformSelectionProxy(self)
        self.cursor = qwave2.WaveformCursorProxy(self)
        self.magAndOutCons = {}   # output controls
        self.waveforms = {}
        self.master = parent
        # flexibles are widgets that are inserted into the second column
        # of the grid.  their width is defined by a resizer waveforms.
        self.flexibles = [self.ruler1,self.ruler2]
        self.resizerWaveform = None
        self.rulerWaveform = None
        self.ignoreWaveformHorizontallyChanged = False
        
    def __getitem__(self, i):
        """
        @return: MyHBox
        """
        if i < 0 or i >= len(self.rows)-2:
            return None
        else:
            return self.rows[i+1]

    def __len__(self):
        return len(self.rows)-2

    def _waveformHorizontallyChanged(self, beg, dur):
        # FIXME: make sure that this flag is protected
        if self.ignoreWaveformHorizontallyChanged: return
        self.ignoreWaveformHorizontallyChanged = True
        for wform in self.waveforms.values():
            s = wform.getBeginSeconds()
            d = wform.getWidthSeconds()
            if s != beg or d != dur:
                wform.display(beg, dur)
        self.ignoreWaveformHorizontallyChanged = False
                
    def resizeFlexibles(self):
        if self.resizerWaveform:
            for row in self.rows:
                if row[1] == self.resizerWaveform:
                    break
            else:
                print "This should not happen!"
                return
            wx = [w.width() for w in row]
            for row in self.rows:
                if row[1] in self.flexibles:
                    row[0].setFixedWidth(wx[0])
                    row[1].setFixedWidth(wx[1])
                    row.setStretchFactor(0,0)
                    row.setStretchFactor(1,0)
                    row.setStretchFactor(2,1)
                    row.setStretchFactor(3,1)
            self.ruler1.redraw()
            self.ruler2.redraw()
            
    def getCellWidth(self, c):
        return self.grid.cellGeometry(0,c).width()
    
    def getSelectionProxy(self):
        return self.selection
    def getCursorProxy(self):
        return self.cursor
    
    def getRowIndex(self, sndfile, ch):
        if (sndfile,ch) not in self.waveforms:
            return None
        wave = self.waveforms[sndfile,ch]
        for i,r in enumerate(self.rows):
            if wave == r[1]:
                return i-1
        return None

    def getWaveform(self, sndfile, ch):
        if (sndfile,ch) not in self.waveforms:
            return None
        else:
            return self.waveforms[sndfile,ch]

    def getWaveforms(self):
        return self.waveforms.values()
    
    def insertRow(self, sndfile, ch, row, waveformOnly=False):
        """
        @param waveformOnly: If True, then vertical ruler, amplitude
        control and output control are not created.
        """
        if (sndfile,ch) in self.waveforms:
            return False
        if row < 0:
            row = 0
        elif row > len(self.rows)-2:
            row = len(self.rows)-2

        hbox = MyHBox(self)
        self.grid.insertWidget(row+1,hbox)

        if len(self.rows) > 2:
            # get any existing waveform
            w0 =  self.rows[1][1] # [vr,w,mag,vol]
            beg = w0.getBeginSeconds()
            dur = w0.getWidthSeconds()
        else:
            # FIXME: user needs to specify these
            beg = 0.0
            dur = 60.0
        wave = self.waveformCls(
            sndfile, ch, beg, dur, self.selection, self.cursor, hbox)

        self.rows.insert(row+1,hbox)

        if waveformOnly:
            hbox.initialize(w2=wave)
            self.grid.setStretchFactor(hbox,0)
            self.flexibles.append(wave)
            if self.resizerWaveform:
                self.resizeFlexibles()
        else:
            vr = qwave2.WaveformVRuler(hbox)
            mag = qt.QSlider(hbox)
            vbox = qt.QVBox(hbox)
            g = qt.QButtonGroup(1,qt.Qt.Vertical,vbox)
            g.setInsideMargin(0)
            g.setInsideSpacing(0)
            chkLeft = qt.QCheckBox(g)
            chkRight = qt.QCheckBox(g)
            vol = qt.QSlider(vbox)

            self.magAndOutCons[sndfile,ch] =  (
                OutputControl(self.master.player, sndfile, ch, vol, chkLeft, chkRight),
                MagControl(mag, wave, vr))

            hbox.initialize(vr,wave,mag,vbox)
            vr.connectToWaveform(wave)

            self.grid.setStretchFactor(hbox,1)
            
            if self.resizerWaveform is None:
                self.resizerWaveform = wave
                self.connect(wave,qt.SIGNAL("waveformWindowResized()"),
                             self.resizeFlexibles)
                
        # connect rulers to the first waveform
        if self.rulerWaveform == None:
            self.ruler1.connectToWaveform(wave)
            self.ruler2.connectToWaveform(wave)
            self.rulerWaveform = wave

        self.connect(wave,qt.SIGNAL("waveformHorizontallyChanged(double,double)"),
                     self._waveformHorizontallyChanged)
        
        self.waveforms[sndfile,ch] = wave
        return True
    
    def removeRow(self, row):
        if row < 0 or row >= len(self.rows) - 2:
            return
        row += 1

        hbox = self.rows[row]
        del self.rows[row]
        
        wform = hbox[1]
        sndfile = wform.getSndFile()
        ch = wform.getChannel()
        del self.waveforms[sndfile,ch]
        self.selection.unregisterWaveform(wform)
        self.cursor.unregisterWaveform(wform)
        if wform in self.flexibles:
            self.flexibles.remove(wform)
        else:
            del self.magAndOutCons[sndfile,ch]
            if wform == self.resizerWaveform:
                for row in self.rows:
                    if row[1] not in self.flexibles:
                        self.resizerWaveform = row[1]
                        self.connect(row[1],
                                     qt.SIGNAL("waveformWindowResized()"),
                                     self.resizeFlexibles)
                        break
                else:
                    self.resizerWaveform = None
                    # there is no resizer waveform
                    # so let the flexibles take all space available
                    for row in self.rows:
                        row[0].setMinimumWidth(0)
                        row[1].setMinimumWidth(0)
                        row[1].setMaximumWidth(10000)
                        row.setStretchFactor(0,0)
                        row.setStretchFactor(1,1)
                        row.setStretchFactor(2,0)
                        row.setStretchFactor(3,0)
        if wform == self.rulerWaveform:
            if len(self.rows) > 2:
                self.rulerWaveform = self.rows[1][1]
                self.ruler1.connectToWaveform(self.rulerWaveform)
                self.ruler2.connectToWaveform(self.rulerWaveform)
            else:
                self.rulerWaveform = None
                
        hbox.close()
        
    def swapRows(self, r1, r2):
        if r1 < 0 or r1 >= len(self.rows)-2 or \
           r2 < 0 or r2 >= len(self.rows)-2:
            return
        r1 += 1
        r2 += 1
        save1 = self.rows[r1]
        save2 = self.rows[r2]
        del self.rows[r1]
        del self.rows[r2]
        self.grid.remove(save1)
        self.grid.remove(save2)
        if r1 < r2:
            self.rows.insert(r1, save2)
            self.rows.insert(r2, save1)
            self.grid.insertWidget(r1, save2)
            self.grid.insertWidget(r2, save1)
        elif r2 < r1:
            self.rows.insert(r2, save1)
            self.rows.insert(r1, save2)
            self.grid.insertWidget(r2, save1)
            self.grid.insertWidget(r1, save2)
        
    def moveRow(self, r1, r2):
        if r1 < 0 or r1 >= len(self.rows)-2:
            return
        if r2 < 0:
            r2 = 0
        elif r2 > len(self.rows)-2:
            r2 = len(self.rows)-2
        r1 += 1
        r2 += 1
        save = self.rows[r1]
        del self.rows[r1]
        self.grid.remove(save)
        if r1 < r2:
            self.rows.insert(r2-1,save)
            self.grid.insertWidget(r2-1,save)
        elif r1 > r2:
            self.rows.insert(r2,save)
            self.grid.insertWidget(r2,save)
        
    def hideRow(self, row):
        if row < 0 or row >= len(self.rows)-2:
            return
        self.rows[row+1].hide()

    def showRow(self, row):
        if row < 0 or row >= len(self.rows)-2:
            return
        self.rows[row+1].show()

    def numRows(self):
        return len(self.rows) - 2

class Waveform(qt.QVBox):
    """ Waveform is an all-in-one class that includes waveforms,
    audio player, and palyback control.
    """

    def __init__(self, parent=None, name=None, waveformCls=WaveformWithResizableSelection):
        qt.QVBox.__init__(self, parent, name)
        self.waveformCls = waveformCls

        ####################
        # UI
        #
        hbox = qt.QHBox(self)
        self.setStretchFactor(hbox,0)
        
        tm = qwave2.TimeLabel(hbox)
        tb = qwave2.TimeLabel(hbox)
        te = qwave2.TimeLabel(hbox)
        td = qwave2.TimeLabel(hbox)
        
        tm.setFrameStyle(qt.QFrame.Box | qt.QFrame.Sunken)
        tb.setFrameStyle(qt.QFrame.Box | qt.QFrame.Sunken)
        te.setFrameStyle(qt.QFrame.Box | qt.QFrame.Sunken)
        td.setFrameStyle(qt.QFrame.Box | qt.QFrame.Sunken)

        tm.setFixedSize(85,20)
        tb.setFixedSize(85,20)
        te.setFixedSize(85,20)
        td.setFixedSize(85,20)

        tm.setAlignment(qt.QLabel.AlignRight)
        tb.setAlignment(qt.QLabel.AlignRight)
        te.setAlignment(qt.QLabel.AlignRight)
        td.setAlignment(qt.QLabel.AlignRight)

        spacer = qt.QWidget(hbox)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding,qt.QSizePolicy.Fixed)

        playBtn = qt.QPushButton(">", hbox)
        repeatBtn = qt.QPushButton("R", hbox)
        pauseBtn = qt.QPushButton("||", hbox)
        stopBtn = qt.QPushButton("X", hbox)

        playBtn.setFixedSize(20,20)
        repeatBtn.setFixedSize(20,20)
        pauseBtn.setFixedSize(20,20)
        stopBtn.setFixedSize(20,20)

        speedSlider = qt.QSlider(hbox)
        speedSlider.setOrientation(qt.QSlider.Horizontal)
        speedSlider.setFixedWidth(60)
        speedSlider.setMinValue(-10)
        speedSlider.setMaxValue(10)

        grid = WaveformGrid(self, self.waveformCls)
        self.setStretchFactor(grid,1)

        sb = qwave2.WaveformScrollBar(self)

        if sys.platform == 'win32':
            player = qwave2.getPlayer(self.winId())
        else:
            player = qwave2.getPlayer()
        player.initialize()
        player.enableTicker()
        ticker = player.getPlayerTicker()

        cursor = grid.getCursorProxy()
        selection = grid.getSelectionProxy()
        selectionColor = qt.QColor(qt.Qt.red)
        unfocusedRegionColor = qt.QColor(qt.Qt.yellow)
        selection.setColorScheme(selectionColor, unfocusedRegionColor)

        
        ticker.registerReceiver(tm)
        ticker.registerReceiver(sb)
        ticker.registerReceiver(cursor)

        ####################
        # behaviour
        #
        self.connect(playBtn, qt.SIGNAL("clicked()"), self.play)
        self.connect(repeatBtn, qt.SIGNAL("clicked()"), self.repeat)
        self.connect(pauseBtn, qt.SIGNAL("clicked()"), self.pauseOrResume)
        self.connect(stopBtn, qt.SIGNAL("clicked()"), self.stop)
        self.connect(speedSlider, qt.SIGNAL("valueChanged(int)"), self.setSpeed)
        self.connect(selection, qt.SIGNAL("waveformSelectionChanged(double,double,Waveform*)"), self.waveformSelectionChanged)

        ####################
        # fields
        #
        self.selectionColor = selectionColor
        self.unfocusedRegionColor = unfocusedRegionColor

        self.grid = grid
        self.sb = sb

        self.tm = tm
        self.tb = tb
        self.te = te
        self.td = td
        self.playBtn = playBtn
        self.repeatBtn = repeatBtn
        self.pauseBtn = pauseBtn
        self.stopBtn = stopBtn
        self.hasRuler = False
        self.sndfiles = {}

        ##
        ## public properties
        self.player = player
        self.cursor = cursor
        self.selection = selection  # the selection proxy
        
    ####################
    # signal emitters
    #
    def waveformSelectionChanged(self, beg, dur, wform):
        """ Process the WaveformSelectionProxy.waveformSelectionChanged
        signal to set labels (TimeLabels) that display the configuration
        of the selection.
        """
        
        self.tb.setTime(beg)
        self.te.setTime(beg+dur)
        self.td.setTime(dur)
        
    ####################
    # player control
    #
    def play(self):
        self.pauseBtn.setText("||")
        beg = self.selection.getBeginSeconds()
        dur = self.selection.getWidthSeconds()
        self.player.play(beg, dur)

    def playRegion(self, beg, dur):
        self.pauseBtn.setText("||")
        self.player.play(beg, dur)
        
    def repeat(self):
        self.pauseBtn.setText("||")
        beg = self.selection.getBeginSeconds()
        dur = self.selection.getWidthSeconds()
        self.player.repeat(beg, dur)
        
    def pauseOrResume(self):
        if self.player.isDevicePlaying():
            self.player.pause()
            self.pauseBtn.setText("=")
        else:
            self.player.resume()
            self.pauseBtn.setText("||")

    def stop(self):
        self.pauseBtn.setText("||")
        self.player.stop()

    def setSpeed(self, r):
        self.player.setSpeed(math.pow(2.0,r/10.0))

    ##############################
    # waveform layout
    #
    def placeWaveform(self, filename, ch, i, waveformOnly=False):
        try:
            sndfile = self.sndfiles[filename]
        except KeyError:
            return

        if self.grid.insertRow(sndfile, ch, i, waveformOnly):
            w = self.grid.getWaveform(sndfile, ch)
            self.sb.registerWaveform(w)
            self.connect(w,qt.SIGNAL("waveformMouseMoved(Waveform*,double)"),
                         self._setCursorLocOnTimeLabel)
            #self.player.setWeight(sndfile, ch, 1.0)
        else:
            row = self.grid.getRowIndex(sndfile, ch)
            self.grid.moveRow(row, i)

    def hideWaveform(self, filename, ch):
        try:
            sndfile = self.sndfiles[filename]
        except KeyError:
            return
        row = self.grid.getRowIndex(sndfile, ch)
        self.grid.hideRow(row)
        #self.player.setWeight(sndfile, ch, 0.0)

    ##############################
    # sndfiles
    #
    def addSndFile(self, filename, addToPlayer=True):
        if filename in self.sndfiles: return
        s = qwave2.SndFile(filename)
        self.player.addSndFile(s)
        for ch in range(0,s.getChannels()):
            self.player.setWeight(s, ch, 0.0)
        self.sndfiles[filename] = s
        return s

    def addDummySndFile(self, sndfile):
        name = str(sndfile)
        if name in self.sndfiles: return
        self.sndfiles[name] = sndfile
        return sndfile
        
    def removeSndFile(self, filename):
        """ Remove the given audio file from the waveform display and player.
        """
        try:
            sndfile = self.sndfiles[filename]
        except KeyError:
            return
        for ch in range(sndfile.getChannels()):
            w = self.grid.getWaveform(sndfile, ch)
            self.sb.unregisterWaveform(w)
            i = self.grid.getRowIndex(sndfile, ch)
            if i is not None:
                self.grid.removeRow(i)
        self.player.stop()
        self.player.removeSndFile(sndfile)
        del self.sndfiles[filename]

    def removeDummySndFile(self, sndfile):
        filename = sndfile.getFileName()
        if filename not in self.sndfiles: return
        for ch in range(sndfile.getChannels()):
            w = self.grid.getWaveform(sndfile, ch)
            self.sb.unregisterWaveform(w)
            i = self.grid.getRowIndex(sndfile, ch)
            if i is not None:
                self.grid.removeRow(i)
        del self.sndfiles[filename]
        
    def getSndFiles(self):
        """
        @return: Dictionary of audio filename and SndFile.
        """
        return copy.copy(self.sndfiles)
    
    ### other
    def getLeftMargin(self):
        return self.grid.getCellWidth(0)
        
    def numWaveforms(self):
        return self.grid.numRows()

    def getMaxEndTime(self):
        t = -9999999.0
        for sndfile in self.sndfiles.values():
            t = max(t,sndfile.getLengthSeconds())
        return t

    def getWaveforms(self):
        return self.grid.getWaveforms()
    
    def getWaveform(self, filename, ch):
        """ Returns the specified waveform.

        @param filename: Audio file name containing the waveform.
        @type filename: str
        @param ch: Channel of the audio file containing the waveform.
        @type ch: int
        """
        if filename in self.sndfiles:
            return self.grid.getWaveform(self.sndfiles[filename], ch)
    
    def getChannels(self, filename):
        if filename in self.sndfiles:
            return self.sndfiles[filename].getChannels()
        else:
            return None
        
    def getSelectedRegionS(self):
        x = self.selection.getBeginSeconds()
        y = x + self.selection.getWidthSeconds()
        wid = self.selection.getSelectedWaveform()
        return (wid, x, y)

    def getCursorPositionS(self):
        return self.cursor.getCursorPositionSeconds()
        
    def setSelectionColorScheme(self, focused=None, unfocused=None):
        """ Set colors for selection and unfocused regions.

        @param focused: The color of the selection.
        @type focused: QColor
        @param unfocused: The color of the unfocused regions.
        @type unfocused: QColor
        """
        if focused is None:
            focused = self.selectionColor
        if unfocused is None:
            unfocused = self.unfocusedRegionColor
        self.selection.setColorScheme(focused, unfocused)
        beg = self.selection.getBeginSeconds()
        dur = self.selection.getWidthSeconds()
        wform = self.selection.getSelectedWaveform()
        self.selection.select(beg, dur, wform)

    def markRegionS(self, filename, ch, beg, end, scroll=True):
        try:
            sndfile = self.sndfiles[filename]
        except KeyError:
            return

        wform = self.grid.getWaveform(sndfile, ch)
        if wform is None: return
        
        s = wform.getBeginSeconds()
        w = wform.getWidthSeconds()
        d = wform.getSecondsPerPixel()

        if scroll:
            if s + d > end:
                s = beg;
            elif s + w - d < beg:
                s = end - w
            for path,sndfile in self.sndfiles.items():
                for ch in range(sndfile.getChannels()):
                    w = self.grid.getWaveform(sndfile, ch)
                    if w is not None:
                        w.display(s)

        self.selection.select(beg, end-beg, wform)
        self.tb.setTime(beg)
        self.te.setTime(end)
        self.td.setTime(end-beg)
            
            
    def zoomInRegion(self, beg, end):
        for wform in self.grid.getWaveforms():
            wform.display(beg, end-beg)

    def zoomInAtCursor(self):
        wforms = self.grid.getWaveforms()
        if wforms:
            a = wforms[0].getBeginSeconds()
            dur = wforms[0].getWidthSeconds() / 2.0
            x = self.cursor.getCursorPositionSeconds()
            beg = a + (x-a)/2.0
            for wform in wforms:
                wform.display(beg, dur)
                #wform.display(beg+dur/2.0)

    def zoomOutAtCursor(self):
        wforms = self.grid.getWaveforms()
        if wforms:
            a = wforms[0].getBeginSeconds()
            dur = wforms[0].getWidthSeconds() * 2.0
            x = self.cursor.getCursorPositionSeconds()
            beg = a - (x-a)
            for wform in wforms:
                wform.display(beg, dur)

    def zoomAtCenter(self, f):
        """ Zoom in/out at the center of the displayed waveform.
        It takes an argument f which is a zoom factor.  If f is larger than
        1.0, the interval of the viewing window is extended, thus the
        waveform is zoomed out.  If f is smaller than 1.0, the viewing
        window gets smaller and the waveform is zoomed in.

        @param f: Zoom factor.
        @type f: float
        """
        wforms = self.grid.getWaveforms()
        if wforms:
            dur = wforms[0].getWidthSeconds()
            dur2 = dur / f
            a = wforms[0].getBeginSeconds() + (dur-dur2)/2.0
            for wform in wforms:
                wform.display(a, dur2)

    def moveCenterAtCursor(self):
        wforms = self.gridWidgets.keys()
        if wforms:
            dur = wforms[0].getWidthSeconds()
            x = self.cursor.getCursorPositionSeconds()
            beg = x - dur/2.0
            for wform in wforms:
                wform.display(beg, dur)

    def moveCenterAt(self, t):
        wforms = self.grid.getWaveforms()
        if wforms:
            dur = wforms[0].getWidthSeconds()
            beg = t - dur/2.0
            wforms[0].display(beg, dur)
            
    def _setCursorLocOnTimeLabel(self, waveform, t):
        """ Used to connect Waveform.waveformMouseMoved and
        TimeLabel.setTime.
        """
        self.tm.setTime(t)
        
if __name__ == "__main__":
    app = qt.QApplication([])
    vbox = qt.QVBox()
    w = Waveform(vbox)
    app.setMainWidget(vbox)
    def f():
        w.hideWaveform("sw.wav", 0)
    b = qt.QPushButton("do something", vbox)
    b.connect(b,qt.SIGNAL("clicked()"),f)
    vbox.show()
    w.addSndFile("sw.wav")
    w.placeWaveform("sw.wav", 0, 0)
    w.placeWaveform("sw.wav", 1, 0)
    app.exec_loop()
    
