from qt import *
from qtcanvas import *
from myaccel import AccelKeyHandler
import speakercode
import bisect

__all__ = ["TranscriptEdit"]


class TranscriptEdit(QTextEdit, AccelKeyHandler):
    def __init__(self, parent=None):
        QTextEdit.__init__(self, parent)
        self.clipboard = QApplication.clipboard()
        self.defaultKeyBindings = {
            "Ctrl+V":"paste"
            }
        self.currentParagraph = None
        self.currentParagraphHeight = None
        self.beginingParagraph = None
        self.setKeyBindings(self.defaultKeyBindings)
        self.data = None
        self.rownums = []
        self.sideBarLayout = []
        self.editActionFiltering = True
        self.sb = self.verticalScrollBar()
        self.sb2 = None # another vertical scroll bar

        # There is a call circle:
        #   data.select -> _select -> (cursor position is changed)
        #     -> _cursorPositionChanged -> data.select
        # To cut the circle, _cursorPositionChanged sets
        # cursorPosAlreadySet to True.  Then, _select doesn't change the
        # cursor position and sets cursorPosAlreadySet to False if
        # it is True.
        self.cursorPosAlreadySet = False

        # to avoid errors that occur in _display
        # (_display tries to disconnect these signals at the beginning)
        self.connect(self, SIGNAL("textChanged()"),self._textChanged)
        self.connect(self, SIGNAL("cursorPositionChanged(int,int)"),self._cursorPositionChanged)
        self.connect(self.sb,SIGNAL("valueChanged(int)"), self._generateSideBarLayout)

    def _display(self, data):
        # these signals have to be disconnected to avoid any side effects
        # caused when changing the text
        self.disconnect(self, SIGNAL("textChanged()"),self._textChanged)
        self.disconnect(self, SIGNAL("cursorPositionChanged(int,int)"),self._cursorPositionChanged)
        self.editActionFiltering = False
        self.clear()
        self.rownums = []
##         i = 0
##         for row in data:
##             if self.displayFilter(row):
##                 self.rownums.append(row.num)
##                 self.insertParagraph(row['transcript'],i)
##                 i += 1
        txt = ""
        for row in data:
            if self.displayFilter(row):
                self.rownums.append(row.num)
                txt += row['transcript'] + "\n"
        if txt:
            self.setText(txt[:-1])
        for i in range(len(self.rownums),self.paragraphs()):
            self.removeParagraph(i)
        if self.rownums:
            data.select(self.rownums[0])
        self.currentParagraph = 0
        self.editActionFiltering = True
        self.connect(self,SIGNAL("textChanged()"),self._textChanged)
        self.connect(self,SIGNAL("cursorPositionChanged(int,int)"),self._cursorPositionChanged)
        self.setCursorPosition(0,0)
        self.emit(SIGNAL("textChanged()"),())
        
    def setData(self, data, displayFilter=lambda x:True):
        self.data = None
        self.displayFilter = displayFilter
        self._display(data)
        self.data = data
        self.connect(data.emitter,PYSIGNAL("cellChanged"),self._cellChanged)
        self.connect(data.emitter,PYSIGNAL("insertRow"),self._insertRow)
        self.connect(data.emitter,PYSIGNAL("takeRow"),self._takeRow)
        self.connect(data.emitter,PYSIGNAL("select"),self._select)
        self._updateScrollBar2()
        self._generateSideBarLayout()
        
    def resizeEvent(self, e):
        QTextEdit.resizeEvent(self, e)
        self._updateScrollBar2()
        self._generateSideBarLayout()
        # FIXME: this is a hack; do we have a better way to do it?
        #self.parent().sidebar.redraw()

    def showEvent(self, e):
        QTextEdit.showEvent(self, e)
        self._updateScrollBar2()
        
    def setFilter(self, displayFilter):
        self.displayFilter = displayFilter
        self._display(self.data)
        self._updateScrollBar2()
        self._generateSideBarLayout()
        
    def getData(self):
        return self.data

    def getSegmentIndex(self, para):
        """
        @param para: paragraph number of the text pane
        @return: segment corresponding to the paragraph
        """
        try:
            return self.rownums[para]
        except IndexError:
            return None

    def getParagraphIndex(self, segid):
        p = bisect.bisect_left(self.rownums, segid)
        if self.rownums[p] == segid:
            return p
        else:
            return None

    def getSideBarLayout(self):
        return self.sideBarLayout
    
    def contentsMousePressEvent(self, e):
        if e.button() == Qt.LeftButton:
            self.beginingParagraph,char = self.getCursorPosition()
        QTextEdit.contentsMousePressEvent(self, e)

    def setKeyBindings(self, keyBindings):
        keyBindings.update(self.defaultKeyBindings)
        AccelKeyHandler.setKeyBindings(self, keyBindings)

    def keyPressEvent(self, e):
        if self.rownums or \
           (e.state()==Qt.ControlButton and \
            (e.key()==Qt.Key_Y or e.key()==Qt.Key_Z)):
            AccelKeyHandler.keyPressEvent(self, e)
        else:
            e.accept()

    def accel_paste(self):
        if self.editActionFiltering:
            s = self.clipboard.text(QClipboard.Clipboard)
            s = unicode(s).split("\n")[0]
            self.clipboard.setText(s, QClipboard.Clipboard)
            self.paste()
        
    def paste(self):
        # since accel_paste takes care of paste action originated
        # from keyboard event, we only need to cover the paste
        # action originated from mouse mid-button click
        if self.editActionFiltering:
            s = self.clipboard.text(QClipboard.Selection)
            s = unicode(s).split("\n")[0]
            self.clipboard.setText(s, QClipboard.Selection)
        QTextEdit.paste(self)

    def cut(self):
        if self.editActionFiltering:
            if unicode(self.selectedText()).find("\n") >= 0:
                return
        QTextEdit.cut(self)

    def removeSelectedText(self, *args):
        if self.editActionFiltering:
            if unicode(self.selectedText()).find("\n") >= 0:
                self.removeSelection()
                return
        QTextEdit.removeSelectedText(self, *args)

    def doKeyboardAction(self, act):
        if self.editActionFiltering:
            if act == self.ActionBackspace:
                if self.getCursorPosition()[1] == 0:
                    return
            elif act == self.ActionDelete:
                p, c = self.getCursorPosition()
                if self.paragraphLength(p) == c:
                    return
            elif act == self.ActionReturn:
                if self.data:
                    self.data.select(self.rownums[self.currentParagraph])
                    self._highlightParagraph(self.currentParagraph)
                    return
            elif act == self.ActionKill:
                p, c = self.getCursorPosition()
                if self.paragraphLength(p) == c:
                    return
            elif act == self.ActionWordBackspace:
                if self.getCursorPosition()[1] == 0:
                    return
            elif act == self.ActionWordDelete:
                p, c = self.getCursorPosition()
                if self.paragraphLength(p) == c:
                    return
        QTextEdit.doKeyboardAction(self, act)

    def setFont(self, f):
        QTextEdit.setFont(self, f)
        self._updateScrollBar2()
        self._generateSideBarLayout()

    def verticalScrollBar2(self, parent):
        """
        Create a separate scrollbar that simulates the behavior of the
        default (vertical) scrollbar.
        """
        if self.sb2 is None:
            sb2 = QScrollBar(parent)
            sb2.connect(sb2,SIGNAL("valueChanged(int)"),self.sb.setValue)
            sb2.connect(self.sb,SIGNAL("valueChanged(int)"),sb2.setValue)
            self.sb2 = sb2
        return self.sb2
            
    # model-to-gui
    def _cellChanged(self, i, c, v, w):
        """
        @param i: row index
        @param c: column index
        @param v: cell value
        @param w: old value
        """
        h = self.data.getColumnName(c)
        if h == 'transcript':
            p = bisect.bisect_left(self.rownums, i)
            if v == self.text(p)[:-1]: return
            self.disconnect(self,SIGNAL("textChanged()"),self._textChanged)
            self.setSelection(p,0,p,self.paragraphLength(p))
            self.insert(v)
            self.connect(self,SIGNAL("textChanged()"),self._textChanged)
            h = self.paragraphRect(p).height()
            if h != self.currentParagraphHeight:
                #self.emit(PYSIGNAL("paragraphHeightChanged"),(p,h))
                self._generateSideBarLayout()
                self.currentParagraphHeight = h
            
        
    def _insertRow(self, i, row):
        if not self.displayFilter(row): return
        p = bisect.bisect_left(self.rownums, i)
        for k in range(p,len(self.rownums)):
            self.rownums[k] += 1
        # self.rownums should be updated before the line is actually
        # inserted into the text pane because when that happens, other
        # components that depend on the list will try to look up the
        # values in the list.  if we haven't updated the list, they
        # will use the old values
        self.rownums.insert(p,i)
        # we are not changing text, but just moving it around
        # so, we don't want the _textChanged get activated
        self.disconnect(self, SIGNAL("textChanged()"),self._textChanged)
        self.disconnect(self,SIGNAL("cursorPositionChanged(int,int)"),self._cursorPositionChanged)
        # note that the newline character triggers the doKeyboardAction()
        # method with "return" action, which in turn calls data.select()
        # so we preset self.currentParagraph so that the waveform display
        # don't jump back and forth when self.setCursorPosition is called
        # at the end
        self.currentParagraph = p
        if p == self.paragraphs():
            self.insertParagraph("\n%s\n" % row['transcript'], p)
        else:
            self.insertParagraph("%s\n" % row['transcript'], p)
        for j in range(len(self.rownums),self.paragraphs()):
            self.removeParagraph(j)
        self.connect(self, SIGNAL("textChanged()"),self._textChanged)
        self.connect(self,SIGNAL("cursorPositionChanged(int,int)"),self._cursorPositionChanged)
        #self.emit(PYSIGNAL("paragraphInserted"), (p,))
        self._generateSideBarLayout()
        self.setCursorPosition(p,0)

    def _takeRow(self, i, r):
        p = bisect.bisect_left(self.rownums, i)
        if i == self.rownums[p]:
            del self.rownums[p]
            for k in range(p,len(self.rownums)):
                self.rownums[k] -= 1
            # we are not changing text, but just moving it around
            # so, we don't want the _textChanged get activated
            self.disconnect(self, SIGNAL("textChanged()"),self._textChanged)
            self.removeParagraph(p)
            self.connect(self, SIGNAL("textChanged()"),self._textChanged)
            #self.emit(PYSIGNAL("paragraphRemoved"), (p,))
            self._generateSideBarLayout()
            if self.rownums:
                if p >= len(self.rownums) and p > 0:
                        #self.setCursorPosition(p-1,0)
                        self.data.select(self.rownums[p-1])
                else:
                    #self.setCursorPosition(p,0)
                    self.data.select(self.rownums[p])
                
    def _select(self, i):
        if self.cursorPosAlreadySet:
            self.cursorPosAlreadySet = False
            return
        if i != self.currentParagraph:
            p = bisect.bisect_left(self.rownums, i)
            _, char = self.getCursorPosition()
            self.cursorPosAlreadySet = True
            self.setCursorPosition(p,char)
            self.currentParagraph = p
            self.currentParagraphHeight = self.paragraphRect(p).height()
            self._highlightParagraph(p)

    # self-to-self
    def _cursorPositionChanged(self, para, char):
        if self.cursorPosAlreadySet:
            self.cursorPosAlreadySet = False
            return
        if self.data and \
           self.currentParagraph != para:
            self.cursorPosAlreadySet = True
            self.data.select(self.rownums[para])
            self.currentParagraph = para
            self.currentParagraphHeight = self.paragraphRect(para).height()
            self._highlightParagraph(para)

    def _textChanged(self):
        if self.data:
            p = self.currentParagraph
            segment = self.data[self.rownums[p]]
            newtext = unicode(self.text(p))[:-1]
            if segment['transcript'] == newtext:
                return
            segment['transcript'] = newtext
            h = self.paragraphRect(p).height()
            if h != self.currentParagraphHeight:
                #self.emit(PYSIGNAL("paragraphHeightChanged"),(p,h))
                self._generateSideBarLayout()
                self.currentParagraphHeight = h

    # utilities
    def _highlightParagraph(self, p):
        self.setSelection(p,0,p,self.paragraphLength(p),777)
        self.setSelectionAttributes(777,Qt.green,False)
        # we do this twice because sometimes the paragraph is highlighted
        # in black.  doing it twice fixes this problem
        self.setSelection(p,0,p,self.paragraphLength(p),777)
        self.setSelectionAttributes(777,Qt.green,False)

    def _updateScrollBar2(self):
        if self.sb2 is not None:
            h = QFontMetrics(self.font()).lineSpacing()
            n = self.lines()
            p = self.viewport().height()
            max = n * h - p + 3 # don't know why have to add 3
            
            self.sb2.setMinValue(0)
            self.sb2.setMaxValue(max)
            self.sb2.setLineStep(20)
            self.sb2.setPageStep(p)
        
    def _generateSideBarLayout(self, *args):
        if not self.data:
            self.sideBarLayout = []
            self.emit(PYSIGNAL("sideBarLayoutChanged"), ([],))
            return

        #paraNum = self.paragraphs()
        paraNum = len(self.rownums)
        p = self.viewportToContents(QPoint(0,0))
        viewportBottom = self.height()
        para = self.paragraphAt(p)
        r = self.paragraphRect(para)
        top = self.contentsToViewport(r.topLeft()).y() + 2
        lineSpacing = QFontMetrics(self.font()).lineSpacing()

        L = []
        while para < paraNum and top < viewportBottom:
            n = self.linesOfParagraph(para)
            height = n * lineSpacing
            seg = self.data[self.rownums[para]]
            L.append((para,seg,top,height))
            top += height
            para += 1
            
        if self.sideBarLayout != L:
            self.sideBarLayout = L
            self.emit(PYSIGNAL("sideBarLayoutChanged"), (L,))

if __name__ == "__main__":
    from transcript_qt import *
    L = [[(None,None)]*10,
         ['sw',0,1.23,2.01,'A','male','native','how are you',1,1,1,'report'],
         ['sw',0,2.01,2.53,'B','female','native',"I'm fine",1,1,2,'report']]
    data = Transcript.importList(L)

    def myFilter(row):
        return row['speakerType'] == 'female'
            
    app = QApplication([])
    w = TranscriptEditWithSideBar()
    #w.setData(data, myFilter)
    app.setMainWidget(w)
    w.show()
    data.insertRow(1,['sw',0,2.01,2.53,'C','female','native',"Okay",1,1,2,'report'])
    app.exec_loop()
    
